// ignore_for_file: public_member_api_docs, invalid_use_of_protected_member

import 'package:flutter/material.dart';

import '../extensions/context.dart';
import 'text_formx_field.dart';

/// A [Exception] thrown by this library.
class FormxException implements Exception {
  /// Default constructor.
  FormxException(this.message) : errorTexts = const {};

  /// Thrown when [FormxState.submit] is called and there are invalid fields.
  FormxException.withErrorTexts(this.errorTexts) {
    final buffer = StringBuffer();
    errorTexts.forEach((tag, msg) => buffer.write('Invalid tag $tag: $msg\n'));
    message = buffer.toString();
  }

  /// The exception message.
  late final String message;

  /// A map of [String] tag and [String] errorText.
  final Map<String, String> errorTexts;

  @override
  String toString() => 'FormxException: $message';
}

/// A widget that builds a form with [TextFormxField] widgets.
class Formx extends StatefulWidget {
  /// A widget that builds a form with [TextFormxField] widgets.
  ///
  /// A form can also be nested in another [Formx] for more complex forms.
  ///
  /// Use `Formx.of(context)` or `GlobalKey` to access the formx state.
  ///
  /// Example:
  ///
  /// ```dart
  /// Formx(
  ///   tag: null,
  ///   child: Formx(
  ///     tag: 'my_nested_form',
  ///     child: TextFormxField('my_field'),
  ///   ),
  /// )
  /// ```
  ///
  /// Which results in the form:
  ///
  /// ```dart
  /// {
  ///  'my_nested_form': {'my_field': 'value'},
  /// }
  const Formx({
    super.key,
    required this.child,
    this.tag,
    this.enabled,
    this.initialValues,
    this.onSaved,
    this.onChanged,
    this.onSubmitted,
    this.valueModifier,
    this.fieldWrapper,
    this.fieldModifier,
    this.errorTextModifier,
    this.decorationModifier,
    this.autovalidateMode,
  });

  /// When present, links this form to a parent [FormxState.form]. This also enables parent/child callbacks.
  /// The tagged form map will be nested in the parent form map.
  ///
  /// Ex: {'my_tag': {'my_field': 'value'}}
  final String? tag;

  /// The widget tree with [TextFormxField] widgets (or [Formx] with [tag]).
  final Widget child;

  /// When false, disables all descendants.
  final bool? enabled;

  /// The initial value for each [TextFormxField.tag] or nested [Formx.tag].
  final FormMap? initialValues;

  /// Listens to all changes of this [FormxState.form].
  final FormCallback? onChanged;

  /// Called when `onFieldSubmitted` or `submit()` are called and valid.
  final FormCallback? onSubmitted;

  /// Called when [FormxState.save] is called.
  final FormCallback? onSaved;

  /// Modifier for each [FormMap] value. [FormxState.form] is built with this.
  final ValueModifier? valueModifier;

  /// Wrapper for each resolved field [Widget].
  final FieldWrapper? fieldWrapper;

  /// Modifier for each [TextFormxField].
  final FieldModifier? fieldModifier;

  /// Modifier for each `TextFormxField.validator` errorText.
  final ErrorTextModifier? errorTextModifier;

  // /Modifier for each `TextFormxField.decoration`.
  final DecorationModifier? decorationModifier;

  /// When null, starts disabled and changes to always on [onSubmitted].
  final AutovalidateMode? autovalidateMode;

  /// Returns the first [FormxState] above [BuildContext].
  static FormxState of(BuildContext context) {
    final maybe = maybeOf(context);
    assert(maybe != null, 'No [Formx] of this context.');
    return maybe!;
  }

  /// Returns the first [FormxState] above [BuildContext] or null.
  static FormxState? maybeOf(BuildContext context) {
    final widget = context
        .getElementForInheritedWidgetOfExactType<_FormxScope>()
        ?.widget as _FormxScope?;

    return widget?.state;
  }

  /// Returns the first [FormxState] below [context]. Filters by [key] or [tag].
  static FormxState at(BuildContext context, {Key? key, String? tag}) {
    return context.visitState(
      assertType: 'Formx',
      filter: (state) =>
          (tag == null || tag == state.widget.tag) &&
          (key == null || key == state.widget.key),
    );
  }

  @override
  State<Formx> createState() => FormxState();
}

/// Map signature for [Formx] form, where:
/// - `key` is the [TextFormxField.tag] or [Formx.tag].
/// - `value` is either a field value [String] or a nested form [FormMap].
typedef FormMap = Map<String, dynamic>;

/// Signature for [FormMap] callback.
typedef FormCallback = void Function(FormMap form);

/// Signature for field wrapper by [TextFormxField.tag].
typedef FieldWrapper = Widget Function(String tag, Widget widget);

/// Signature for field modifier by [TextFormxField.tag].
typedef FieldModifier = TextFormxField Function(
  String tag,
  TextFormxField field,
);

/// Signature for form value modifier by [Formx.tag].
typedef ValueModifier = dynamic Function(String tag, dynamic value);

/// Signature for error text modifier by [TextFormxField.tag].
typedef ErrorTextModifier = String Function(String tag, String errorText);

/// Signature for decoration modifier by [TextFormxField.tag].
typedef DecorationModifier = InputDecoration? Function(
  String tag,
  InputDecoration? decoration,
);

/// The state of a [Formx] widget.
///
/// You can access the state using [Formx.key], [Formx.of] or [Formx.at].
class FormxState extends State<Formx> {
  /// The parent [FormxState] descendant of this [context].
  late final FormxState? parent = () {
    final scope = Formx.maybeOf(context);
    if (!hasTag) return scope;
    assert(
      hasTag && scope != null,
      'This [Formx] has a tag but no parent was found. '
      'When [tag] is present, you must wrap this form '
      'in a parent [Formx] widget',
    );
    scope?.nested[widget.tag!] = this;
    return scope;
  }();

  /// Whether this [Formx] has a tag.
  bool get hasTag => widget.tag != null;

  // * Formx inheritance
  bool? get enabled => widget.enabled ?? (hasTag ? parent?.enabled : null);
  FieldWrapper? get fieldWrapper => widget.fieldWrapper ?? parent?.fieldWrapper;
  ValueModifier? get valueModifier =>
      widget.valueModifier ?? parent?.valueModifier;
  FieldModifier? get fieldModifier =>
      widget.fieldModifier ?? parent?.fieldModifier;
  ErrorTextModifier? get errorTextModifier =>
      widget.errorTextModifier ?? parent?.errorTextModifier;
  DecorationModifier? get decorationModifier =>
      widget.decorationModifier ?? parent?.decorationModifier;
  AutovalidateMode? get autovalidateMode =>
      widget.autovalidateMode ?? (hasTag ? parent?.autovalidateMode : null);

  /// The initial values for this [form]. Applied to [fields] and [nested].
  late final FormMap initialValues = widget.initialValues != null
      ? Map.of(widget.initialValues!)
      : Map.of(parent?.initialValues[widget.tag] as FormMap? ?? {});

  /// The nested [FormxState] immediate children, keyed by [Formx.tag].
  @protected
  final nested = <String, FormxState>{};

  /// The [FormFieldState] immediate fields, keyed by [TextFormxField.tag].
  @protected
  final fields = <String, FormFieldState<String>>{};

  /// Returns a flat [Map] of all [fields] and [nested] errorTexts.
  Map<String, String> get errorTexts {
    final errorTexts = <String, String>{};
    fields.forEach((tag, field) => errorTexts[tag] = field.errorText ?? '');
    nested.forEach((tag, formx) => errorTexts.addAll(formx.errorTexts));
    return errorTexts..removeWhere((key, value) => value.isEmpty);
  }

  /// Returns a [FormMap] if [validate] succeeds, or throws a [FormxException].
  FormMap submit() {
    return trySubmit() ?? (throw FormxException.withErrorTexts(errorTexts));
  }

  /// Returns a [FormMap] if [validate] succeeds, or null.
  FormMap? trySubmit() {
    if (!validate()) return null;
    widget.onSubmitted?.call(form);
    return form;
  }

  /// Returns a [FormMap] with all [fields] and [nested] values.
  ///
  /// If [valueModifier] is present, it will be applied to each value.
  FormMap get form {
    var form = <String, dynamic>{};

    void maybeModify(String tag, dynamic value) {
      if (valueModifier != null) {
        form[tag] = valueModifier!(tag, value);
      } else {
        form[tag] = value;
      }
    }

    fields.forEach((tag, field) => maybeModify(tag, field.value));
    nested.forEach((tag, formx) => maybeModify(tag, formx.form));

    return form;
  }

  /// Returns true if [fields] or [nested] have errors. Doesn't set ui state.
  bool get hasError => [
        fields.values.any((field) => field.hasError),
        nested.values.any((formx) => formx.hasError),
      ].any((hasError) => hasError);

  /// Returns true if [fields] and [nested] are valid. Doesn't set ui state.
  bool get isValid => [
        fields.values.every((field) => field.isValid),
        nested.values.every((formx) => formx.isValid),
      ].every((isValid) => isValid);

  /// Validates all [fields] and [nested]. Sets ui state.
  bool validate() {
    fields.forEach((tag, field) => field.validate());
    nested.forEach((tag, formx) => formx.validate());
    return !hasError;
  }

  /// Fills this [FormxState.form] with new [values].
  ///
  /// Values that are not [String] or [FormMap] will be ignored.
  void fill(FormMap values) {
    values.forEach((tag, value) {
      if (value is FormMap) nested[tag]?.fill(value);
      if (value is String) setValue(tag, value);
    });
  }

  /// Resets all [fields] and [nested] to their initial values.
  void reset() {
    fields.forEach((tag, field) => field.reset());
    nested.forEach((tag, formx) => formx.reset());
  }

  /// Saves this [form], all [fields] and [nested] fields.
  void save() {
    fields.forEach((tag, field) => field.save());
    nested.forEach((tag, formx) => formx.save());
    widget.onSaved?.call(form);
  }

  /// Sets the value of the field with the enclosing [tag].
  ///
  /// If [notify] is false, the field widget will not be rebuilt.
  void setValue(String tag, String value, {bool notify = true}) {
    if (notify) {
      fields[tag]?.didChange(value);
    } else {
      fields[tag]?.setValue(value);
    }
  }

  /// Attaches a [FormFieldState] to this [FormxState].
  @protected
  void attachField(String tag, FormFieldState<String> field) {
    fields[tag] = field;
  }

  /// Reports a change to this [FormxState.form].
  @protected
  void onChanged(String tag, String value) {
    widget.onChanged?.call(form);

    //If tag is present, callbacks to parent.
    if (hasTag) parent?.widget.onChanged?.call(parent!.form);
  }

  /// Reports a submit to this [FormxState.form].
  @protected
  void onSubmitted(String tag, String value) {
    if (!validate()) return;
    widget.onSubmitted?.call(form);

    //If tag is present, callbacks to parent.
    if (hasTag) parent?.widget.onSubmitted?.call(parent!.form);
  }

  @override
  Widget build(BuildContext context) {
    return _FormxScope(
      state: this,
      child: widget.child,
    );
  }
}

class _FormxScope extends InheritedWidget {
  const _FormxScope({required super.child, required this.state});
  final FormxState state;

  @override
  bool updateShouldNotify(_FormxScope oldWidget) => false;
}
